from django.template import TemplateSyntaxError, NodeList, Template

# TODO: Find out why Django raises InvalidTemplateLibrary without this import.
from django.template.loader import get_template

from django.template.loader_tags import BlockNode, ExtendsNode
from django.template.response import SimpleTemplateResponse

from djpj.compat import queue

_wrapped_class_registry = {}


class StopRendering(Exception):
    """
    Thrown in DjPjNodeList.render() when all template blocks have been found
    and rendered, to stop rendering the rest of the page, which saves a bit of
    template rendering time.
    """
    pass


class DjPjObject(object):
    """
    This is a base class used for wrapping various Django template structures.
    This works by dynamically changing the type of various objects, so beware!

    The purpose of DjPjObject is essentially to monkey-patch arbitrary objects
    in a very visible way, making use of Python's class system instead of
    overriding instance methods for transparent debugging and IDE-friendliness.

    To use, define your patched methods on a subclass of DjPjObject, then
    pass the object you want to patch to patch(). The object's type will be
    swapped to one that includes your patch class in its class hierarchy, and
    Python's method resolution machinery will call the patched methods instead
    of their original counterparts. You can use self and super in your methods
    as if you were subclassing normally.

    Optionally, you can include additional base classes. In that case, you'll
    see an error if you try to patch an object that is not an instance of
    every extra base class.
    """

    def __new__(cls, *args, **kwargs):
        raise NotImplementedError("%s cannot be instantiated directly. "
                                  "Use the patch() method instead." % cls)

    def __patch__(self, *args, **kwargs):
        """
        This is kind of like __init__, but run after an object is patched,
        rather than after it's created. Override it to provide patch-time
        behaviour.
        """
        pass

    @classmethod
    def patch(cls, obj, *args, **kwargs):
        """
        Create (or fetch from cache) a new class that inherits from both cls
        and the passed object's class.
        """
        if not isinstance(obj, cls):
            for base_class in cls.__bases__[1:]:
                if not isinstance(obj, base_class):
                    raise RuntimeError(
                        "Object to be patched by %s is not an instance of %s"
                        % (cls.__name__, base_class.__name__))
            obj_class = type(obj)
            try:
                new_class = _wrapped_class_registry[obj_class]
            except KeyError:
                new_class = type("DjPj" + obj_class.__name__,
                                 (cls, obj_class), {})
                _wrapped_class_registry[obj_class] = new_class
            obj.__class__ = new_class
            obj.__patch__(*args, **kwargs)
        return obj


class DjPjNodeList(DjPjObject, NodeList):
    """
    When using the pjax_block decorator, the nodelists of matching BlockNodes
    are set to this class. We override the NodeList instead of BlockNode
    because during the rendering process, an entirely new BlockNode is created
    and rendered (see BlockNode.render() in django/template/loader_tags.py),
    bypassing any overridden behaviour in our subclass.
    """

    def __patch__(self, block_name):
        self._djpj_block_name = block_name

    def render(self, context):
        result = super(DjPjNodeList, self).render(context)
        try:
            if self._djpj_block_name in context.djpj_blocks:
                context.djpj_blocks[self._djpj_block_name] = result
                if None not in context.djpj_blocks.values():
                    raise StopRendering
        except AttributeError:
            pass
        return result


class DjPjExtendsNode(DjPjObject, ExtendsNode):
    def get_parent(self, *args, **kwargs):
        parent = super(DjPjExtendsNode, self).get_parent(*args, **kwargs)
        DjPjTemplate.patch(parent)
        return parent


class DjPjTemplate(DjPjObject, Template):
    """
    Use this by casting your template with DjPjTemplate.patch(template), and
    then calling template.render_blocks(context, blocks), where blocks is a
    sequence of block names you wish to render.
    """

    def __patch__(self):
        self._djpj_initialised_blocks = set()
        self._initialise_blocks()

    def _initialise_blocks(self):
        """
        Walk the template tree and convert all necessary objects into their
        DjPj equivalents. This includes ExtendsNodes and BlockNodes' NodeLists
        (but not BlockNodes themselves as they're not actually rendered from
        the template tree - see the source for BlockNode.render().
        """
        node_queue = queue.Queue()
        node_queue.put(self)
        while True:
            try:
                node = node_queue.get(False)
            except queue.Empty:
                break
            if hasattr(node, 'nodelist'):
                if isinstance(node, BlockNode):
                    DjPjNodeList.patch(node.nodelist, node.name)
                    self._djpj_initialised_blocks.add(node.name)
                for child_node in node.nodelist:
                    node_queue.put(child_node)
            if isinstance(node, ExtendsNode):
                DjPjExtendsNode.patch(node)
        del node_queue

    def render_blocks(self, context, blocks):
        """
        Return a dict mapping block names to their rendered contents. If a
        block is not rendered, its name will map to None.
        """
        context.djpj_blocks = dict((b, None) for b in blocks if b)
        try:
            self.render(context)
        except StopRendering:
            pass
        return context.djpj_blocks


class PJAXTemplateResponse(DjPjObject, SimpleTemplateResponse):
    """
    This is used by the PJAX decorator. Before a response is returned, this
    class is inserted into its type hierarchy, so that we can intercept calls
    to rendered_content() and instead
    it's inserted into its type hierarchy, so that when rendered_content is
    accessed, we
    """

    def __patch__(self, block_name, title_block_name, title_variable):
        self._djpj_block_name = block_name
        self._djpj_title_block_name = title_block_name
        self._djpj_title_variable = title_variable

    @property
    def rendered_content(self):
        """
        Walk the template's node tree, casting our target blocks' nodelists to
        PJAXBlockNodeList in order to store its output in the render context.
        Then render the template as usual, but instead of returning the result,
        return the captured output from our target block(s).
        """

        # just for convenience
        block = self._djpj_block_name
        title_block = self._djpj_title_block_name
        title_var = self._djpj_title_variable

        # Get a Template and Context object
        template = self.resolve_template(self.template_name)
        context = self.resolve_context(self.context_data)

        # If no block name is specified, assume we're rendering a PJAX-specific
        # template and just return the rendered output.
        if not block:
            return template.render(context)

        # Otherwise, proceed to capture the output from the pjax block and,
        # if specified, the title block or variable.
        DjPjTemplate.patch(template)
        target_blocks = filter(None, (block, title_block))
        rendered_blocks = template.render_blocks(context, target_blocks)

        # Get all our error handling out of the way before generating
        # our PJAX-friendly output
        if None in rendered_blocks.values():
            raise TemplateSyntaxError("Template block '%s' does not exist or was not rendered" % block)
        if title_var and title_var not in context:
            raise KeyError("PJAX title variable '%s' not found in context" % title_var)

        # Return our PJAX response including a <title> tag if necessary
        block_contents = rendered_blocks[block]
        title_contents = rendered_blocks.get(title_block, None) or context.get(title_var)
        title_html = "<title>%s</title>\n" % title_contents if title_contents else ""

        return title_html + block_contents



import ast
import re

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from djpj.decorator import pjax_block, pjax_template
from djpj.utils import strip_pjax_parameter


class DjangoPJAXMiddleware(object):

    def __init__(self, config=None):
        djpj_setting = (config or getattr(settings, 'DJPJ_PJAX_URLS', None))
        self.decorated_urls = self.parse_configuration(djpj_setting or [])

    @staticmethod
    def parse_decorator(decorator_string):

        error = lambda msg: ImproperlyConfigured(
            '"%s" does not define a valid PJAX decorator: %s'
            % (decorator_string, msg))

        if not decorator_string.startswith('@'):
            raise error("expression should start with '@'")

        expr = ast.parse(decorator_string[1:], '<string>', mode='eval')
        if not isinstance(expr.body, ast.Call):
            raise error("decorator expression must be a single call "
                        "to pjax_block or pjax_template")

        call = expr.body
        if call.func.id not in ('pjax_block', 'pjax_template'):
            raise error("decorator expression must be a single call "
                        "to pjax_block or pjax_template")

        if not (call.starargs is None and call.kwargs is None):
            raise error("unpacking * and ** arguments is not supported")

        if not all(isinstance(arg, ast.Str) for arg
                   in call.args + [kw.value for kw in call.keywords]):
            raise error("only string arguments are allowed")

        return eval(compile(expr, '<string>', mode='eval'))

    @staticmethod
    def parse_configuration(config_tuple):
        listify = lambda d: d if isinstance(d, (list, tuple)) else [d]
        parse_fn = DjangoPJAXMiddleware.parse_decorator
        return tuple(
            (re.compile(url_regex),
             [parse_fn(d) for d in reversed(listify(decorators))])
            for url_regex, decorators in reversed(config_tuple))

    def process_request(self, request):
        strip_pjax_parameter(request)

    def process_template_response(self, request, response):
        for url_regex, decorators in self.decorated_urls:
            if url_regex.match(request.path):
                fake_view = lambda _: response
                for decorator in decorators:
                    response = decorator(fake_view)(request)
        return response



import functools

from django.http import HttpResponseRedirect, HttpRequest
from djpj.template import PJAXTemplateResponse
from djpj.utils import (strip_pjax_parameter, is_pjax,
                        pjax_container, pjaxify_template_var_with_container)


def _make_decorator(partition_fn, process_fn):

    # Import this here to avoid import issues when running tests.
    from django.views.decorators.vary import vary_on_headers

    def djpj_decorator(view):
        @functools.wraps(view)
        def wrapped_view(request, *args, **kwargs):
            response = view(request, *args, **kwargs)
            if partition_fn(request):
                # Before generating a response, strip the "_pjax" GET parameter
                # that jquery-pjax adds as a browser cache-busting measure.
                strip_pjax_parameter(request)

                # This header helps jquery-pjax correctly handle redirects.
                response['X-PJAX-URL'] = (response.get('Location')
                                          or request.get_full_path())
                # Test if response supports deferred rendering, approach copied
                # from django.core.handlers.base.BaseHandler.get_response()
                if hasattr(response, 'render') and callable(response.render):
                    process_fn(request, response)
                elif not isinstance(response, HttpResponseRedirect):
                    raise TypeError("PJAX views must return either a response "
                                    "with a render() method, or a redirect.")
            return response
        return vary_on_headers('X-PJAX-Container')(wrapped_view)

    return djpj_decorator

_make_pjax_decorator = functools.partial(_make_decorator, is_pjax)
_make_ajax_decorator = functools.partial(_make_decorator, HttpRequest.is_ajax)


def pjax_template(template=pjaxify_template_var_with_container):

    if not template:
        raise ValueError("The template argument to pjax_template "
                         "may not be None.")

    def process_response(request, response):
        _template = (template(request, response.template_name)
                     if callable(template) else template)
        if not _template:
            raise ValueError("Tried to set PJAX response's template to %s. "
                             "You must provide a template!" % _template)
        response.template_name = _template

    return _make_pjax_decorator(process_response)


def pjax_block(block=pjax_container,
               title_variable=None, title_block=None):

    if not block:
        raise ValueError("The block argument to pjax_block may not be None.")

    if title_variable and title_block:
        raise ValueError("Only one of 'title_variable' and 'title_block' "
                         "may be passed to pjax decorator.")

    def process_response(request, response):
        _block = block(request) if callable(block) else block
        PJAXTemplateResponse.patch(response, _block,
                                  title_block, title_variable)

    return _make_pjax_decorator(process_response)



from contextlib import contextmanager
import re


_container_re = re.compile(r'^#\S+$')


def pjax_container(request):
    container = request.META['HTTP_X_PJAX_CONTAINER']
    if _container_re.match(container):
        return container[1:]
    else:
        raise ValueError("Invalid PJAX selector '%s' found in request: "
                         "must be a simple ID selector of the form #<id>."
                         % container)


def pjaxify_template_path(template_path, container):
    try:
        parts = template_path.rsplit('.', 1)
    except AttributeError:
        raise ValueError("template_path must be a string type")
    pjax_identifier = "=".join(filter(None, ("pjax", container)))
    return ".".join(["%s-%s" % (parts[0], pjax_identifier)] + parts[1:])


def transform_template_var(transform_fn, template_var, container=None):
    if not isinstance(template_var, (list, tuple)):
        template_var = (template_var,)
    template_pair = lambda name: (transform_fn(name, container), name)
    return type(template_var)(t for name in template_var
                              for t in template_pair(name))


def pjaxify_template_var(request, template_var):
    return transform_template_var(pjaxify_template_path, template_var)


def pjaxify_template_var_with_container(request, template_var):
    return transform_template_var(pjaxify_template_path, template_var,
                                  container=pjax_container(request))


def is_pjax(request):
    return 'HTTP_X_PJAX' in request.META


@contextmanager
def mutable_querydict(querydict):
    initially_mutable = querydict._mutable
    querydict._mutable = True
    yield querydict
    querydict._mutable = initially_mutable


def strip_pjax_qs_parameter(url):
        return re.sub(r'_pjax=[^&]+&?', '', url).rstrip('&')


def strip_pjax_parameter(request):
    """
    The _pjax GET parameter helps browsers with caching, but is
    unnecessary with the presence of the X-PJAX-Container header,
    and can cause trouble with code that doesn't expect it, so let's
    just pretend it never existed.
    """
    if is_pjax(request):
        if '_pjax' in request.GET:
            with mutable_querydict(request.GET) as get:
                del get['_pjax']
            request.META['QUERY_STRING'] = \
                strip_pjax_qs_parameter(request.META['QUERY_STRING'])



import os
try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='DjPj',
    version='0.3.1',
    description='A template-block-based Django helper for jQuery-PJAX.',
    license='BSD',
    long_description=read('README.rst').replace('__ #', '__ https://pypi.python.org/pypi/djpj/#'),
    url='https://github.com/alexhill/djpj',
    author='Alex Hill',
    author_email='alex@hill.net.au',

    packages=find_packages(),
    install_requires=['django>=1.4'],

    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Programming Language :: Python',
    ],
)
